#include <stdio.h>
#include <stdarg.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h> /* String function definitions */
#include <fcntl.h> /* File control definitions */
#include <errno.h> /* Error number definitions */

struct speed_spec {
	char *name;
	int flag;
};

struct node {
	int x;
	struct node *next;
	struct node *prev;
};

struct speed_spec speeds[] = { { "1200", B1200 }, { "2400", B2400 }, { "4800",
		B4800 }, { "9600", B9600 }, { "19200", B19200 }, { "38400", B38400 }, {
		"57600", B57600 }, { "115200", B115200 }, { "230400", B230400 }, { NULL,
		0 } };

int bufferSize = 512;

struct node *sockfds = NULL;
int sock_change_read = 0;
int sock_change_write = 0;
pthread_mutex_t sock_mutex;
pthread_cond_t change_cond;
pthread_attr_t thread_attr;

char *serial_device_name = NULL;
char *option_baud_rate = NULL;
int option_debug_level = 0;

int listen_port = -1;
int serialfd = -1;
int listenSocketFd = -1;

int finished = 0;

/************************
 *  FUNCTION PROTOTYPES *
 ************************/

int main(int argc, char *argv[]);
/*Thread rutines*/
void *readSerial(void *ptr);
void *writeSerial(void *ptr);
void *acceptSocket(void *ptr);

/*Utility functions*/
int parse_args(int argc, char * argv[]);
int skip_param(char *buf);
void show_help(const char *appName);
int prefix(const char *pre, const char *str);
int get_baud(char * szbaud);
void log_message(char *msg, int level, ...);

/*Initialization functions*/
int init_listen_socket_fd();

void set_thread_attributes(pthread_attr_t *thread_attr);
void setTerminalAttr(int serialfd, long BAUD);

void list_add_first(struct node **root, int value);
void list_remove_node(struct node **root, struct node *rem);
int list_remove_value(struct node **root, int value);
int list_length(struct node *root);

/*****************
 * MAIN PROGRAM  *
 *****************/

int main(int argc, char *argv[]) {
	pthread_t read_thread, write_thread, accept_thread;
	long BAUD;

	/*Parse arguments and derive baudrate*/
	parse_args(argc, argv);
	BAUD = get_baud(option_baud_rate);

	/* Check serial name*/
	if (!prefix("/dev/tty", serial_device_name)) {
		if (prefix("tty", serial_device_name)) {
			char temp[20] = "/dev/";
			strcat(temp, serial_device_name);
			serial_device_name = temp;
		} else {
			log_message("%s is not correct serial port name.", 0,
					serial_device_name);
			exit(EXIT_FAILURE);
		}
	}

	/* Open the serial port d*/
	serialfd = open(serial_device_name, O_RDWR);
	if (serialfd < 0) {
		log_message("ERROR open device %s", 0, serial_device_name);
		exit(EXIT_FAILURE);
	}
	setTerminalAttr(serialfd, BAUD);

	/* Connect to the socket server */
	if (!init_listen_socket_fd()) {
		close(serialfd);
		log_message("FAIL to open %d listen port.", 0, listen_port);
		exit(EXIT_FAILURE);
	}

//	/*Accept first socket*/
//	int newsockfd = -1;
//	while (newsockfd < 0) {
//		newsockfd = accept(listenSocketFd, NULL, NULL );
//	}
//	log_message("New socket accepted with FD:%d", 1, newsockfd);
//	list_add_first(&sockfds, newsockfd);

	/*Init Mutexes*/
	pthread_mutex_init(&sock_mutex, NULL );
	pthread_cond_init(&change_cond, NULL );
	pthread_attr_init(&thread_attr);

	set_thread_attributes(&thread_attr);

	/* create reading and writing threads */
	pthread_create(&read_thread, NULL, readSerial, NULL );
	pthread_create(&write_thread, NULL, writeSerial, NULL );
	pthread_create(&accept_thread, NULL, acceptSocket, NULL );
	log_message("Threads started.", 1);
	/* Wait for the threads to finish */
	pthread_join(read_thread, NULL );
	log_message("Read thread stopped by join", 1);
	pthread_join(write_thread, NULL );
	log_message("Write thread stopped by join", 1);
	close(listenSocketFd);
	pthread_join(accept_thread, NULL );
	log_message("Accept thread stopped by join", 1);

	/* Clean mutexes*/
	pthread_mutex_destroy(&sock_mutex);
	pthread_cond_destroy(&change_cond);

	/*Close all accepted sockets*/
	while (sockfds != NULL ) {
		log_message("Sockte FD:%d closed", 2, sockfds->x);
		close(sockfds->x);
		sockfds = sockfds->next;
	}

	close(serialfd);
	log_message("Finished!", 0);
	return 0;
}

int parse_args(int argc, char * argv[]) {

	char **loc_argv = argv;
	int loc_argc = argc;
	int skip;

	while (loc_argc > 1) {
		if (loc_argv[1][0] == '-')
			switch (loc_argv[1][1]) {
			case 'h':
				show_help(argv[0]);
				exit(0);
				break;
			case 'p':
				skip = skip_param(&loc_argv[1][0]);
				listen_port = atoi(&loc_argv[1][skip]);
				log_message("ARG, listen port=%d", 1, listen_port);
				break;
			case 's':
				skip = skip_param(&loc_argv[1][0]);
				serial_device_name = &loc_argv[1][skip];
				log_message("ARG, serial port=%s", 1, serial_device_name);
				break;
			case 'b':
				skip = skip_param(&loc_argv[1][0]);
				option_baud_rate = &loc_argv[1][skip];
				log_message("ARG, baud rate=%d", 1, option_baud_rate);
				break;
			case 'g':
				skip = skip_param(&loc_argv[1][0]);
				option_debug_level = atoi(&loc_argv[1][skip]);
				log_message("ARG, debug level=%d", 1, option_debug_level);
				break;

			default:
				show_help(argv[0]);
				log_message("ERROR Wrong argument: %s", 0, loc_argv[1]);
				exit(EXIT_FAILURE);
			}

		++loc_argv;
		--loc_argc;
	}

	if (serial_device_name == 0 || listen_port < 0) {
		show_help(argv[0]);
		log_message("Error missing serial device name or port number", 0);
		exit(EXIT_FAILURE);
	}
	return EXIT_SUCCESS;
}

void show_help(const char *appName) {
	fprintf(stdout,
	"Usage: %s -p <socket listen port> -s <serial port dev>\n\n"
	"  -h, -help                 display this help and exit\n"
	"  -p port                   socket port to listen on\n"
	"  -s <serial device>        serial device ex /dev/ttyS1\n"
	"options\n"
	"  -b baudrate               set buad rate default to 9600\n"
	"  -g                        debug level 0-3\n"
	"\n", appName);
}

int skip_param(char *buf) {
	int x = 0;
	if (buf[2] != 0)
		return 2;
	/* skip any other whitespace */
	x = 3;
	while (buf[x] == ' ' && x < 40)
		x++;
	if (x < 40)
		return x;
	return 0;
}

void *readSerial(void *ptr) {
	char c[bufferSize];
	int n;
	int *fds = NULL;
	int *fds_ptr = NULL;
	int fds_length = 0;
	struct node *rp_ptr = NULL;
	int i;

	sock_change_read = 1;

	log_message("Start read thread.", 1);

	while (finished == 0) {
		/* read from the serial port */
		n = read(serialfd, &c, sizeof(c));
		if (n < 0) {
			log_message("ERROR reading from serial port", 0);
			finished = 1;
			continue;
		}

		/* Print data from serial data*/
		if (option_debug_level > 2) {
			printf("SERIAL>");
			for (i = 0; i < n; ++i) {
				printf("[%02x]", c[i]);
			}
			printf("\n");
		}

		/*Check change on sockfds list*/
		pthread_mutex_lock(&sock_mutex);
		if (sock_change_read) {
			log_message("Update socket array.", 2);
			/*Update fds array*/
			fds_length = list_length(sockfds);

			/*Wait for socket*/
			if (fds_length < 1) {
				log_message("No socket available for writing, data lost.", 2);
//				pthread_cond_wait(&change_cond, &sock_mutex);
//				fds_length = list_length(sockfds);
//				log_message("%d new sockets ready", 2, fds_length);
				pthread_mutex_unlock(&sock_mutex);
				continue;
			}

			free(fds);
			fds = malloc(fds_length * sizeof(int));

			/*Write LinkedList state to fds array*/
			fds_ptr = fds;
			rp_ptr = sockfds;

			log_message("Socket fds: ", 2);
			while (rp_ptr != NULL ) {
				log_message(">>>%d", 2, rp_ptr->x);
				*fds_ptr = rp_ptr->x;
				fds_ptr++;
				rp_ptr = rp_ptr->next;
			}
			sock_change_read = 0;
		}
		pthread_mutex_unlock(&sock_mutex);

		fds_ptr = fds;
		for (i = 0; i < fds_length; i++) {
			/* write to the socket */

			n = write(*fds_ptr, &c, n);
			log_message("%d bytes written to FD: %d", 2, n, *fds_ptr);
			if (n < 0) {
				log_message("ERROR writing to socket:%d", 0, *fds_ptr);
				pthread_mutex_lock(&sock_mutex);
				list_remove_value(&sockfds, *fds_ptr);
				sock_change_read = 1;
				sock_change_write = 1;
				pthread_mutex_unlock(&sock_mutex);
			}
			fds_ptr++;
		}
	}
	free(fds);
	return NULL ;
}

void *writeSerial(void *ptr) {
	char c[bufferSize];
	int ready;
	int n;
	fd_set rfds;
	fd_set read_rfds;
	int max_fd;
	struct timeval tv;
	struct node *rp_ptr = NULL;

	int i;

	sock_change_write = 1;

	log_message("Start write thread.", 0);

	while (finished == 0) {
		/*Check change on sockfds list*/
		pthread_mutex_lock(&sock_mutex);
		if (sock_change_write) {
			if (list_length(sockfds) < 1) {
				/*No socket opened*/
				log_message("No sockets available for read, wait...", 2);
				pthread_cond_wait(&change_cond, &sock_mutex);
				log_message("New sockets available for read.", 2);
			}

			FD_ZERO(&rfds);
			max_fd = -1;
			rp_ptr = sockfds;
			log_message("FD_SET:", 2);
			while (rp_ptr != NULL ) {
				log_message(">>>>%d", 2, rp_ptr->x);
				FD_SET(rp_ptr->x, &rfds);
				if (max_fd < rp_ptr->x) {
					max_fd = rp_ptr->x;
				}
				rp_ptr = rp_ptr->next;
			}
			sock_change_write = 0;
		}
		pthread_mutex_unlock(&sock_mutex);

		/*Select readable socket*/

		tv.tv_sec = 5;
		tv.tv_usec = 0;
		FD_ZERO(&read_rfds);
		memcpy(&read_rfds, &rfds, sizeof(fd_set));
		ready = select(max_fd + 1, &read_rfds, NULL, NULL, &tv);

		if (ready < 1) {
			log_message("select(): timeout on %d", 2, ready);
			continue;
		}

		for (i = 0; i <= max_fd; i++) {
			if (FD_ISSET(i, &read_rfds)) {
				log_message("Socket FD:%d ready", 2, i);
				n = recv(i, &c, sizeof(c), MSG_DONTWAIT);
				if (n == 0) {
					log_message("ERROR reading from socket %d", 0, n);
					pthread_mutex_lock(&sock_mutex);
					list_remove_value(&sockfds, i);
					sock_change_read = 1;
					sock_change_write = 1;
					pthread_mutex_unlock(&sock_mutex);
				} else if (n < 0) {
					log_message("Socket %d timed out!", 1, i);
				} else {
					/* Print data from serial data*/
					if (option_debug_level > 2) {
						printf("SOCKET>");
						for (i = 0; i < n; ++i) {
							printf("[%02x]", c[i]);
						}
						printf("\n");
					}
					n = write(serialfd, &c, n);
					log_message("%d bytes written to serial,", 2, n);
					if (n < 0) {
						log_message("ERROR writing to serial", 0);
						finished = 1;
						break;
					}
				}

			}
		}

	}
	return NULL ;
}

void *acceptSocket(void *ptr) {
	int newsockfd;

	while (finished == 0) {
		newsockfd = accept(listenSocketFd, NULL, NULL );
		log_message("New socket accepted with FD:%d", 1, newsockfd);
		if (newsockfd > 0) {
			struct timeval tv;
			tv.tv_sec = 1;
			tv.tv_usec = 0;
//			setsockopt(newsockfd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
			pthread_mutex_lock(&sock_mutex);
			list_add_first(&sockfds, newsockfd);
			sock_change_read = 1;
			sock_change_write = 1;
			log_message("Socket added FD:%d", 2, newsockfd);
			pthread_cond_signal(&change_cond);
			pthread_mutex_unlock(&sock_mutex);
		} else {
			log_message("ERROR Listen socket failed", 0);
			finished = 1;
			close(serialfd);
		}
	}
	return NULL ;
}

int init_listen_socket_fd() {
	int bOptionTrue = 1;
	struct sockaddr_in serv_addr;

	/* create a listening socket fd */
	listenSocketFd = socket(AF_INET, SOCK_STREAM, 0);
	if (listenSocketFd < 0) {
		log_message("ERROR creating our listening socket", 0);
		return (0);
	}

	/* clear our socket address structure */
	bzero((char *) &serv_addr, sizeof(serv_addr));

	serv_addr.sin_addr.s_addr = INADDR_ANY;
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(listen_port);

	setsockopt(listenSocketFd, SOL_SOCKET, SO_REUSEADDR, (char *) &bOptionTrue,
			sizeof(bOptionTrue));

	if (bind(listenSocketFd, (struct sockaddr *) &serv_addr, sizeof(serv_addr))
			< 0) {
		log_message("ERROR binding to server port", 0);
		return (0);
	}

	listen(listenSocketFd, 10);

	log_message("Listening socket created on port %i", 1, listen_port);

	return 1;
}

int prefix(const char *pre, const char *str) {
	return strncmp(pre, str, strlen(pre)) == 0;
}

int get_baud(char * szbaud) {
	struct speed_spec *s;
	int speed = 0;
	if (szbaud != 0) {
		for (s = speeds; s->name; s++) {
			if (strcmp(s->name, szbaud) == 0) {
				speed = s->flag;
				break;
			}
		}
	}
	/* default to 7 in our array or 115 200 */
	if (speed == 0)
		s = &speeds[7];
	log_message("Setting speed %s\n", 1, s->name);
	return s->flag;
}

void set_thread_attributes(pthread_attr_t *thread_attr) {
	return;
}

void setTerminalAttr(int serialfd, long BAUD) {
	struct termios tattr1;
	// Set terminal attributes
	tcgetattr(serialfd, &tattr1);
	tattr1.c_lflag &= ~ICANON; // Disable erase, kill, werase, and rprnt special characters
	tattr1.c_lflag &= ~ECHO; // Dont echo input characters
	tattr1.c_lflag &= ~ISIG; // Disable interrupt, quit, and suspend special characters
	tattr1.c_lflag &= (~ECHONL & ~ECHOK); // Dont echo new line
	tattr1.c_iflag &= ~ICRNL; // Dont translate carriage return to new line
	tattr1.c_iflag &= ~IGNBRK; // Dont ignore break characters
	tattr1.c_iflag &= ~IXON; // Disable XON/XOFF flow control
	tattr1.c_oflag &= ~ONLCR; //Disable NewLine mapping

	tattr1.c_cflag &= ~CSTOPB; // Use only one stop bit

	tattr1.c_cflag |= CREAD; // Allow input to be received

	tattr1.c_cc[VMIN] = 1; // Minimal number of 1 received bytes
	tattr1.c_cc[VTIME] = 1; // Timeout between received bytes in seconds 10^-1

	// Set the input/output speed
	cfsetispeed(&tattr1, BAUD);
	cfsetospeed(&tattr1, BAUD);
	tcsetattr(serialfd, TCSAFLUSH, &tattr1);
}

void log_message(char *msg, int level, ...) {
	static char message[1024];
	if (level <= option_debug_level) {
		va_list args;
		va_start(args, level);
		vsnprintf(message, 1024, msg, args);
		fprintf(stdout, "%s\n", message);
		fflush(stdout);
		va_end(args);
	}
}

/*****************************
 * List manipulation methods *
 *****************************/
void list_add_first(struct node **root, int value) {
	struct node *rp = malloc(sizeof(struct node));
	rp->x = value;
	rp->next = *root;
	if (*root != NULL ) {
		(*root)->prev = rp;
	}
	rp->prev = NULL;
	*root = rp;
}

void list_remove_node(struct node **root, struct node *rem) {
	if (rem == NULL ) {
		return;
	}

	if (*root == rem) {
		*root = rem->next;
		rem->next->prev = NULL;
		return;
	}

	(rem->prev)->next = rem->next;
	if (rem->next != NULL ) {
		(rem->next)->prev = rem->prev;
	}
}

int list_remove_value(struct node **root, int value) {
	struct node *rp = *root;
	int removed = 0;
	while (rp != NULL ) {
		if (rp->x == value) {
			list_remove_node(root, rp);
			struct node *rp_rem = rp;
			rp = rp->next;
			free(rp_rem);
			removed++;
		} else {
			rp = rp->next;
		}
	}
	return 0;
}

int list_length(struct node *root) {
	if (root == NULL ) {
		return 0;
	}
	return (1 + list_length(root->next));
}
